// =====================================================================================================================
// bmcremedydataload.js
//
// Main file for the BMC Remedy data load integration service. Handles requests for batch and dynamic loading of
// user, group and team data from Remedy to xMatters.
//
// Remedy data is accessed via the xm-ctm-XXX.js files that wrap xMatters custom Web services in Remedy.
//
// Data sync include / exclude lists are defined in dataSync.js
//
// =====================================================================================================================
importPackage(java.util);
importPackage(java.io);
importPackage(java.text);

importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLMessage);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLMessageImpl);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLToken);
importClass(Packages.com.alarmpoint.integrationagent.security.EncryptionUtils);
importClass(Packages.com.alarmpoint.integrationagent.soap.exception.SOAPRequestException);

importClass(Packages.com.thoughtworks.xstream.converters.reflection.PureJavaReflectionProvider);

importClass(Packages.org.apache.commons.httpclient.Header);
importClass(Packages.org.apache.commons.httpclient.HttpVersion);
importClass(Packages.org.mule.providers.http.HttpResponse);

load("integrationservices/bmcremedy/lib/javascript/core/baseclass.js");
load("integrationservices/bmcremedy/lib/javascript/core/logger.js");
load("integrationservices/bmcremedy/lib/javascript/webservices/wsutil.js");
load("integrationservices/bmcremedy/lib/javascript/webservices/soapfault.js");
load("integrationservices/bmcremedy/lib/javascript/xmatters/xmattersws.js");

// Core Javascript files for data sync
load("integrationservices/bmcremedy/lib/javascript/xmatters/xMattersEventWS.js");
load("integrationservices/bmcremedy/lib/javascript/xmatters/xMattersUserSync.js");
load("integrationservices/bmcremedy/lib/javascript/xmatters/xMattersGroupSync.js");
load("integrationservices/bmcremedy/lib/javascript/xmatters/dataSyncHelper.js");
load("integrationservices/bmcremedy/lib/javascript/xmatters/dataSyncDevice.js");
load("integrationservices/bmcremedy/lib/javascript/xmatters/dataSyncUser.js");
load("integrationservices/bmcremedy/lib/javascript/xmatters/dataSyncGroup.js");
load("integrationservices/bmcremedy/lib/javascript/xmatters/dataSyncGroupMember.js");
load("integrationservices/bmcremedy/lib/javascript/xmatters/dataSyncTeam.js");
load("integrationservices/bmcremedy/lib/javascript/xmatters/dataSyncCoverage.js");

// Files shared by the different Remedy integration services
load("integrationservices/bmcremedy/util.js");

// Integration-specific Javascript files
load("integrationservices/bmcremedy/bmcremedydataload/configuration.js");
load("integrationservices/bmcremedy/bmcremedydataload/dataSyncList.js");
load("integrationservices/bmcremedy/bmcremedydataload/xm-ctm-people-ws.js");
load("integrationservices/bmcremedy/bmcremedydataload/xm-ctm-support-group-ws.js");
load("integrationservices/bmcremedy/bmcremedydataload/xm-ctm-support-group-assoc-ws.js");
load("integrationservices/bmcremedy/bmcremedydataload/xm-ctm-support-group-functional-role-ws.js");
load("integrationservices/bmcremedy/bmcremedydataload/bmcremedydataload-lifecycle.js");
load("integrationservices/bmcremedy/bmcremedydataload/processUsers.js");
load("integrationservices/bmcremedy/bmcremedydataload/processGroups.js");
load("integrationservices/bmcremedy/bmcremedydataload/phonenumber.js");
load("integrationservices/bmcremedy/bmcremedydataload/userRolesMapper.js");
load("integrationservices/bmcremedy/bmcremedydataload/supervisorMapper.js");

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * apia_http
 *
 * This is the entry point for HTTP requests from Remedy that initiate data load operations.
 *
 * This method expects that the POST body contains a SOAP 1.1 message indicating the data load action to be taken.
 *
 * ---------------------------------------------------------------------------------------------------------------------
 */
function apia_http(httpRequestProperties, httpResponse)
{
    log.debug("Enter - apia_http");
    log.debug("httpRequestProperties: " + httpRequestProperties);

    var httpRequest = null;
    var requestBody = null;
    var responseBody = null;

    try
    {
        httpRequest = httpRequestProperties.getProperty("http.request");
        requestBody = new XML(httpRequestProperties.getProperty("REQUEST_BODY"));
//      log.debug("Handling request to: " + httpRequest);

        log.info("Request body: " + requestBody);

        // -------------------------------------------------------------------------------------------------------------
        // Requests from Remedy are not processed immediately by this code but are put in a queue object and will
        // be handled asynchronously by queueConsumer function below running in another thread.
        //
        // This allows batch data load operations to run for longer than the HTTP timeout mechanism would allow.
        // -------------------------------------------------------------------------------------------------------------
        var obj = new Object();

        obj.action = requestBody.*::Body.*::TriggerRequest.*::action;
        obj.id = requestBody.*::Body.*::TriggerRequest.*::id;

        if (isEmpty(obj.action))
        {
            throw "apia_http: Invalid request - no action element found."
        }

        // Add the request to the BlockingQueue for processing by the dataload thread
        log.debug("apia_http: queuing action [" + obj.action + "], id [" + obj.id + "]");
        queue.put(obj);

        responseBody = makeSoapResponseBody(APIA_HTTP_RECEIVED, "Request was submitted for processing");
    }
    catch (e)
    {
        responseBody = makeSoapResponseBody(APIA_HTTP_ERROR, e.toString());
    }
    finally
    {
        httpResponse.setStatusLine(HttpVersion.HTTP_1_1, 200);
        httpResponse.setHeader(new Header("Content-Type", httpRequestProperties.getProperty("Content-Type")));
        httpResponse.setBodyString(responseBody);
    }

    log.debug("apia_http - response body [" + responseBody + "]");
    log.debug("Exit - apia_http");

    return httpResponse;
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * queueConsumer
 *
 * Main function consuming request objects from the asynchronous queue. Retrieves items from the request queue and
 * passes them on for processing.
 *
 * The queue object and keepThreadAlive and shouldContinue flags are defined in bmcremedydataload-lifecycle.js
 *
 * queue.take() will block and wait until an item is added to the queue by the code above.
 *
 * ---------------------------------------------------------------------------------------------------------------------
 */
function queueConsumer()
{
    log.debug("Enter - queueConsumer");

    while (keepThreadAlive)
    {
        if (shouldContinue)
        {
            try
            {
                // Reset the success/warning/failure collections
                failedList = new ArrayList();
                successList = new ArrayList();
                warningList = new ArrayList();
                reprocessedList = new ArrayList();
                
                var queueItem = queue.take();
                processDataLoadRequest(queueItem);

                logSyncResults(queueItem);
            }
            catch (e)
            {
                log.debug("queueConsumer - Exception: " + e.toString());
            }
        }
    }

    log.debug("Exit - queueConsumer");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * processDataLoadRequest
 *
 * Data loading will be carried out by the library classes in integrationservices/bmcremedy/lib/javascript/xmatters,
 * subject to the configuration settings in dataSyncList.js
 *
 * ---------------------------------------------------------------------------------------------------------------------
 */
function processDataLoadRequest(requestObject)
{
    log.debug("Enter - processDataLoadRequest");

    log.info("processDataLoadRequest: processing request - action [" + requestObject.action + "], id [" + requestObject.id + "]");

    var isUserAction = false;
    var isGroupAction = false;
    var isUpdateOnly;

    // For add, update and delete actions the format of the id determines whether this request applies to
    // a person or a group
    if (!isEmpty(requestObject.id))
    {
        isUserAction = requestObject.id.toLowerCase().indexOf("ppl") != -1;
        isGroupAction = requestObject.id.toLowerCase().indexOf("sgp") != -1;
    }

    switch (String(requestObject.action.toLowerCase()))
    {
        case ACTION_LOAD:
            log.info("processDataLoadRequest: starting batch data load");

            processRemedyUsers(XM_CTM_PEOPLE_WS_BATCH_QUALIFICATION, PROCESS_ADD_UPDATE, ALLOW_ADD_USER);
            processRemedySupportGroups(XM_CTM_SUPPORT_GROUP_WS_BATCH_QUALIFICATION, PROCESS_ADD_UPDATE);
            break;

        case ACTION_ADD:
        case ACTION_UPDATE:
            if (isUserAction)
            {
                // A user that doesn't match the general selection criteria should only be updated
                isUpdateOnly = !isQualifyingUser(requestObject.id);
                processRemedyUsers(XM_CTM_PEOPLE_WS_PERSON_QUALIFICATION.replace("${PERSON_ID}", String(requestObject.id)), PROCESS_ADD_UPDATE, isUpdateOnly);
            }
            else if (isGroupAction)
            {
                // A group that doesn't match the general selection criteria should only be updated
                isUpdateOnly = !isQualifyingGroup(requestObject.id);
                processRemedySupportGroups(XM_CTM_SUPPORT_GROUP_WS_GROUP_QUALIFICATION.replace("${GROUP_ID}", String(requestObject.id)), PROCESS_ADD_UPDATE, isUpdateOnly);
            }
            else
            {
                throw("Request action and id could not be matched to an operation")
            }
            break;

        case ACTION_DELETE:
            if (isUserAction)
            {
                processRemedyUsers(XM_CTM_PEOPLE_WS_PERSON_QUALIFICATION.replace("${PERSON_ID}", String(requestObject.id)), PROCESS_DELETE);
            }
            else if (isGroupAction)
            {
                processRemedySupportGroups(XM_CTM_SUPPORT_GROUP_WS_GROUP_QUALIFICATION.replace("${GROUP_ID}", String(requestObject.id)), PROCESS_DELETE);
            }
            else
            {
                throw("Request action and id could not be matched to an operation")
            }
            break;


        default:
            throw("Request action and id could not be matched to an operation")
    }

    log.debug("Exit - processDataLoadRequest");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * logSyncResults
 *
 * Logs the results from the synchronization process.
 *
 * Also sends a QuickMessage for batch loads if this option is enabled or if there were any warnings or errors.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function logSyncResults(syncRequest)
{
    log.debug("Enter - logSyncResults");

    var syncAction = syncRequest.action + (isEmpty(syncRequest.id) ? "" : ", Remedy ID: " + syncRequest.id);
    var totalCount = successList.size() + warningList.size() + failedList.size();
    var successCount = successList.size();
    var warningCount = warningList.size();
    var failCount = failedList.size();


    var msg = "Synchronization Summary\n\n";
    msg += "Action: " + syncAction + "\n";
    msg += "Total Number of Objects Processed: " + totalCount + "\n";
    msg += "Successful Count: " + successCount + "\n";
    msg += "Successful with Warnings Count: " + warningCount + "\n";
    msg += "Failure Count: " + failCount + "\n\n";

    log.warn(msg);

    if (log.isDebugEnabled())
    {
        if (successList.size() > 0)
        {
            log.debug("Synchronization Success Summary:");
            for (var i = 0; i < successList.size(); i++)
            {
                log.debug(successList.get(i));
            }
        }
    }

    if (warningList.size() > 0)
    {
        log.warn("Synchronization Success with Warnings Summary:");
        for (var i = 0; i < warningList.size(); i++)
        {
            log.warn(warningList.get(i));
        }
    }

    if (failedList.size() > 0)
    {
        log.warn("Synchronization Failure Summary:");
        for (var i = 0; i < failedList.size(); i++)
        {
            log.warn(failedList.get(i));
        }
    }

    if ((SEND_SYNC_SUMMARY && (syncRequest.action == ACTION_LOAD))
        || (warningList.size() > 0)
        || (failedList.size() > 0))
    {
        sendSyncSummaryMessage(syncAction, totalCount, successCount, warningCount, failCount);
    }

    log.debug("Exit - logSyncResults");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * sendSyncSummaryMessage
 *
 * Sends a message to xMatters with a summary of the sync results
 *
 * @param msg String containing the message to send
 * ---------------------------------------------------------------------------------------------------------------------
 */
function sendSyncSummaryMessage(syncAction, totalCount, successCount, warningCount, failCount)
{
    log.debug("Enter - sendSyncSummaryMessage");

    var apxml = ServiceAPI.createAPXML();

    apxml.setMethod("ADD");
    apxml.setSubclass("action");
    apxml.setToken("recipients", XMATTERS_ADMINISTRATOR);
 
    apxml.setToken("action", syncAction);

    // use toString() on the numeric count values to avoid them being formatted as floating point when they get to xMatters

    apxml.setToken("total_count", totalCount.toString());
    apxml.setToken("success_count", successCount.toString());
    apxml.setToken("warning_count", warningCount.toString());
    apxml.setToken("error_count", failCount.toString());

    apxml.setToken("has_warnings", warningCount > 0 ? "true" : "false");
    apxml.setToken("has_errors", failCount > 0 ? "true" : "false");

    var notes = "For further information, please refer to the Integration Agent log file";
    apxml.setToken("notes", notes);

    try
    {
        ServiceAPI.sendAPXML(apxml);
        log.debug("Finished sending apxml to xMatters");
    }
    catch (e)
    {
        log.error("Caught exception processing [sync summary] Exception:" + e);
    }

    log.debug("Exit - sendSyncSummaryMessage");
}


/**
 * ---------------------------------------------------------------------------------------------------------------------
 * apia_response
 *
 * This is the main method for handling APXML messages sent to the
 * Integration Service by AlarmPoint.  The messages may be responses to
 * previous submissions, or they may be requests to the Integration Service
 * that originate from AlarmPoint
 * (e.g., via an ExternalServiceMessage/ExternalServiceRequest2).
 *
 * Any APXMLMessage object that this method returns will be sent to AlarmPoint
 * either via the Integration Service's outbound queues or directly as a
 * Web Service response, depending on the mechanism used to submit the APXML.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function apia_response(apxml)
{
    log.debug("Enter - apia_response");
    log.debug("apia_response - method [" + apxml.getMethod() + "], transaction ID [" + apxml.getTransactionId() + "]");

    /**
     * Only need to handle OK and Error responses from xMatters.
     */
    var method = apxml.getMethod();

    if (APXMLMessage.OK_METHOD.equalsIgnoreCase(method))
    {
        return handleOK(apxml);
    }
    else if (APXMLMessage.ERROR_METHOD.equalsIgnoreCase(method))
    {
        return handleERROR(apxml);
    }
    else if (APXMLMessage.REQUEST_METHOD.equalsIgnoreCase(method))
    {
        return handleRequest(apxml);
    }
    else if (APXMLMessage.SEND_METHOD.equalsIgnoreCase(method))
    {
        return handleSend(apxml);
    }
    else
    {
        throw new IllegalArgumentException("Unrecognized APXML method: " + method);
    }

    log.debug("Exit - apia_response");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * handleOK
 *
 * OK responses are returned by AlarmPoint for each APXML message that it accepts
 * from an Integration Service.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function handleOK(apxml)
{
    var id = apxml.getTransactionId();
    log.info("Submission of transaction " + id + " succeeded.");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * handleERROR
 *
 * ERROR responses are returned by AlarmPoint for each APXML message that it
 * is unable to accept from an Integration Service.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function handleERROR(apxml)
{
    var id = apxml.getTransactionId();
    var code = apxml.getValue(APXMLMessage.ERROR_CODE);
    log.error("Submission of transaction " + id + " failed with error code " + code);
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * Requests are forwarded to the Send handler. Messages may come from ESM or ESR to the IA.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function handleRequest(apxml)
{
    log.debug("handleRequest received apxml, forwarding to handleSend");
    return handleSend(apxml);
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * The main routine for handling Requests from AlarmPoint
 *
 * The Data Load integration does not push any information about data load notifications back to Remedy so just log it.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function handleSend(apxml)
{
    var id = apxml.getTransactionId();
    log.info("Response from xMatters for transaction " + id + " received.");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * This is the main method for handling APXML messages sent to the
 * Integration Service by APClient.  The messages may be requests to perform
 * local activity, or they may be requests to make submissions to AlarmPoint.
 * <p>
 * Any APXMLMessage object that this method returns will be sent to xMatters
 * via the Integration Service's outbound queues.
 * ---------------------------------------------------------------------------------------------------------------------
 */
//function apia_input(apxml)
//{
//    // Forward APXML to xMatters.
//    return apxml;
//}

