/**
 * ---------------------------------------------------------------------------------------------------------------------
 * processRemedySupportGroups
 *
 * Update or delete in xMatters a support group and group association / membership data from Remedy.
 * Group list information is retrieved in chunks whose size is configurable, in order not to retrieve
 * hundreds or user records in one Web service invocation.
 *
 * The delete action needs to be handled by this function because its necessary to retrieve the group information
 * from Remedy to determine the name of the group.
 *
 * Depending on the qualification this may process a single group or multiple groups.
 *
 * ---------------------------------------------------------------------------------------------------------------------
 */
function processRemedySupportGroups(qualification, isDelete, isUpdateOnly)
{
    log.debug("Enter - processRemedySupportGroups");

    log.info("processRemedySupportGroups - isDelete [" + isDelete + "], qualification [" + qualification + "], chunkSize [" + XM_CTM_WS_CHUNK_SIZE + "], isUpdateOnly [" + isUpdateOnly + "]");

    var supportGroupListWS = new XM_CTM_SUPPORT_GROUP_WS(new WSUtil(), XM_CTM_SUPPORT_GROUP_WS_URL);
    var supportGroupList = null;
    var startRecord = 0;

    var userDataSyncHelper = new DataSyncHelper(syncList, "users");

    // Retrieve the support group list from Remedy in chunks until we get a null response.

    while ((supportGroupList = supportGroupListWS.getSupportGroupList(qualification, startRecord, XM_CTM_WS_CHUNK_SIZE)) != null)
    {
        // Get the list of nodes that contain the group descriptions
        log.debug("supportGroupList length [" + supportGroupList.length + "]");

        for each (var supportGroup in supportGroupList)
        {
            var supportGroupId = supportGroup.Support_Group_ID;

            log.debug("startRecord [" + startRecord + "], Remedy Support Group ID [" + supportGroupId + "]");

            tryProcessOneRemedySupportGroup(supportGroup, userDataSyncHelper, isDelete, isUpdateOnly);
            startRecord++;
        }
    }

    log.debug("Exit - processRemedySupportGroups");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * isQualifyingGroup
 *
 * Determine whether the group identified meets the qualifications for batch updates.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function isQualifyingGroup(groupId)
{
    log.debug("Enter - isQualifyingGroup");
    log.debug("isQualifyingGroup: groupId [" + groupId + "]");

    var result = false;

    // Build a Remedy qualification that uses both the ID of this person and the batch processing criteria
    var qualification = XM_CTM_SUPPORT_GROUP_WS_GROUP_QUALIFICATION.replace("${GROUP_ID}", String(groupId))
        + " AND ( "
        + XM_CTM_SUPPORT_GROUP_WS_BATCH_QUALIFICATION
        + " )";

    log.debug("isQualifyingGroup: qualification [" + qualification + "]");

    var groupListWS = new XM_CTM_SUPPORT_GROUP_WS(new WSUtil(), XM_CTM_SUPPORT_GROUP_WS_URL);

    var groupList = groupListWS.getSupportGroupList(qualification, 0, XM_CTM_WS_CHUNK_SIZE);

    if (groupList == null)
    {
        log.debug("isQualifyingGroup: group does not match batch selection criteria");
    }
    else
    {
        if (groupList.length == 1)
        {
            result = true;
            log.debug("isQualifyingGroup: group matches batch selection criteria");
        }
        else
        {
            log.warn("isQualifyingGroup: groupList contains more than one entry");
        }
    }

    log.debug("isQualifyingGroup: result [" + result + "]");
    log.debug("Exit - isQualifyingGroup");

    return result;
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * tryProcessOneRemedySupportGroup
 *
 * Add/update or delete a single Remedy support group in xMatters. If anything goes wrong at this level, record it in the
 * failure list, but continue processing
 * ---------------------------------------------------------------------------------------------------------------------
 */
function tryProcessOneRemedySupportGroup(remedySupportGroup, userDataSyncHelper, isDelete, isUpdateOnly)
{
    log.debug("Enter - tryProcessOneRemedySupportGroup");
    var group = null;

    try
    {
        if (isDelete)
        {
            group = makeGroupForDelete(remedySupportGroup);
        }
        else
        {
            // Get the list of users who are members of this support group
            var supportGroupId = remedySupportGroup.Support_Group_ID;
            var supportGroupMemberList =
                (new XM_CTM_SUPPORT_GROUP_ASSOC_WS(new WSUtil(), XM_CTM_SUPPORT_GROUP_ASSOC_WS_URL)).getGroupMemberList(supportGroupId);

            group = makeGroupForAddUpdate(remedySupportGroup, supportGroupMemberList, userDataSyncHelper, isUpdateOnly);
        }

        // syncGroup will do a delete if the 2nd argument is true
        (new xMattersGroupSync(syncList)).syncGroup(group, isDelete);

        log.info("Finished processing for Remedy group " + remedySupportGroup.Support_Group_ID);
    }
    catch (e)
    {
        var groupName = buildXmattersGroupName(remedySupportGroup.Company, remedySupportGroup.Support_Organization, remedySupportGroup.Support_Group_Name);
        log.warn("tryProcessOneRemedySupportGroup - while processing user Remedy Login ID [" + groupName + "] got exception [" + e.toString() + "]");
        failedList.add((isDelete ? "Delete" : "Add/Update") + " Group: " + groupName + " failed to synchronize: " + e.toString());
    }

    log.debug("Exit - tryProcessOneRemedySupportGroup");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * makeGroupForDelete
 *
 * Delete a single group from xMatters, using the group description from Remedy.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function makeGroupForDelete(remedySupportGroup)
{
    log.debug("Enter - makeGroupForDelete");

    var group = new Group(buildXmattersGroupName(remedySupportGroup.Company, remedySupportGroup.Support_Organization, remedySupportGroup.Support_Group_Name));

    log.info("Deleting group from xMatters [" + group.name + "]");
    log.debug("Exit - makeGroupForDelete");
    return group;
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * makeGroupForAddUpdate
 *
 * Load a single Remedy support group and its associated team / membership information.
 *
 * ---------------------------------------------------------------------------------------------------------------------
 */
function makeGroupForAddUpdate(remedySupportGroup, supportGroupMemberList, userDataSyncHelper, isUpdateOnly)
{
    log.debug("Enter - makeGroupForAddUpdate");
    log.info("Starting data load for Remedy Support Group " + remedySupportGroup.Support_Group_ID + ", isUpdateOnly [" + isUpdateOnly + "]");

    log.debug("remedySupportGroup: " + remedySupportGroup);
    log.debug("supportGroupMemberList: " + supportGroupMemberList);

    var group = new Group(buildXmattersGroupName(remedySupportGroup.Company, remedySupportGroup.Support_Organization, remedySupportGroup.Support_Group_Name));

    group.updateOnly = isUpdateOnly;

    // Some group properties are not derived from Remedy information. These are only given initial values when the group
    // is created. After that, the values in xMatters are preserved during updates.
    if (!group.groupExists)
    {
        group.externallyOwned = EXTERNALLY_OWN_GROUPS;

        // The default supervisor is defined in configuration.js and depends on whether xMatters has
        // merged administrator accounts or not.
        group.supervisors = (new SupervisorMapper()).getSupervisorsForGroup(group);
    }

    // The remainder of the user properties are based on Remedy information at the time of the update

    group.active = remedySupportGroup.Status == "Enabled";
    group.description = remedySupportGroup.Description;

    group.coverage.groupName = group.name;
    group.team.groupName = group.name;
    group.team.externallyOwned = group.externallyOwned;

    if (supportGroupMemberList != null && supportGroupMemberList.length > 0)
    {
        for each (var remedyGroupMember in supportGroupMemberList)
        {
            var userLoginId = remedyGroupMember.Login_ID;
            log.debug("userLoginId [" + userLoginId + "]");

            if (userDataSyncHelper.checkIncludeExcludeList("user", userLoginId))
            {
                syncGroupMember = new GroupMember();
                syncGroupMember.targetName = userLoginId;
                group.team.members.put(userLoginId, syncGroupMember);
            }
        }
    }

    log.debug("About to sync Group: " + group);
    log.debug("Exit - makeGroupForAddUpdate");

    return group;
}


